/**************************************************************************
*SAMSUNG 11Mbps Wireless LAN MiniPCI/PCI device driver for Linux (swld12).
* 
* Copyright (c) 2000~2002 SAMSUNG Electro-Mechanics CO.,Ltd. -- http://www.MagicLan.com
* All rights reserved.
*  
*--------------------------------------------------------------------------
* This driver supports the following features:
*   - Access Point and Ad-Hoc (peer-to-peer) communication 
*   - Wired Equivalent Privacy (WEP) 
*   - Driver utility interface (UIL)
*---------------------------------------------------------------------------
* 
* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * ALTERNATIVELY, this driver may be distributed undger the terms of
 * the following license, in which case the provisions of this license
 * are required INSTEAD OF the GNU General Public License. (This clause
 * is necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.

*HISTORY

*/

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/kernel.h>
//#include <linux/sched.h>
//#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>


////#ifdef HAS_WIRELESS_EXTENSIONS
#include <linux/wireless.h>
//#if WIRELESS_EXT < 5
//#error "Wireless extension v5 or newer required"
//#endif
////#endif

#include "swld12.h"
#include "semdebug_cs.h"
#include "swlioctl.h"

/********************************************************************
 * MISC
 */

//#if (LINUX_VERSION_CODE < VERSION(2,3,99))
// typedef struct device NET_DEVICE;
//#else
 typedef struct net_device NET_DEVICE;
//#endif

//Driver Version
#define SEM_DRIVER_MAJOR_VERSION 0
#define SEM_DRIVER_MINOR_VERSION 2
#define SEM_DRIVER_VARIANT_VERSION 0

static char *version = "0.21";

// Module parameters
//static u_int irq_mask = 0xdeb8;			            // Interrupt mask(IRQ 3,4,5,7,9,10,11,12,14,15)
//static int irq_list[4] = { -1 };			           // Interrupt list (alternative)
static int networkmode = 1;				           // 0:802.11 Ad_hoc 1: Infrastructure, 3: Ad-hoc
static char stationname[ESSID_MAX_SIZE+1] = "\0";       // Name of station []
static char essid[ESSID_MAX_SIZE+1] = "any\0";	      // Name of ESSID
//static int Domain = 48;                               // 16 : FCC/IC(ch1-11), 48 :Europe,Korea,Japan(Ch1-13),								
								                // 49 : Spain(ch10-11), 50 : France(ch10-13)
static int txrate = 0x0f;			           //Transmit rate [15]
static int psmode = 0;                                  //0: disabled , 1: Enabled.
static int wepreq = 0;                                  //0: disabled,  3: 64 bit WEP, 7: 128bit WEP.
static int keyid=0;                                     // Default WEP KEY ID.
#define MAX_KEY_LEN (2+(13*2))
static char* key0= "";//[MAX_KEY_LEN+1]= "\0";
static char* key1= "";//[MAX_KEY_LEN+1]= "\0";
static char* key2="";//[MAX_KEY_LEN+1]= "\0";
static char* key3="";//[MAX_KEY_LEN+1]= "\0";
static int channel = 4;				           // Channel [4]
static int apdensity = 1;				      // AP density [1]
static int threshold = 2347;			           // RTS threshold [2347]
static int fragthresh = 2346;			           // Fragmentation threshold [2346]

static int mtu = 1500;
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");

//MODULE_PARM(irq_mask, "i");
//MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(networkmode, "i");
MODULE_PARM(stationname, "c" __MODULE_STRING(ESSID_MAX_SIZE));
MODULE_PARM(essid, "c" __MODULE_STRING(ESSID_MAX_SIZE));
//MODULE_PARM(Domain, "i");
MODULE_PARM(psmode, "i");
MODULE_PARM(wepreq, "i");
MODULE_PARM(keyid, "i");
MODULE_PARM(key0, "s");//"c" __MODULE_STRING(MAX_KEY_LEN));
MODULE_PARM(key1, "s");//"c" __MODULE_STRING(MAX_KEY_LEN));
MODULE_PARM(key2, "s");//"c" __MODULE_STRING(MAX_KEY_LEN));
MODULE_PARM(key3, "s");//"c" __MODULE_STRING(MAX_KEY_LEN));

MODULE_PARM(channel, "i");
MODULE_PARM(apdensity, "i");
MODULE_PARM(threshold, "i");
MODULE_PARM(fragthresh, "i");
MODULE_PARM(txrate, "i");
MODULE_PARM(mtu, "i");

// Ethernet timeout is ((400*HZ)/1000), but we should use a higher
// value, because wireless transmissions are much slower
#define TX_TIMEOUT ((4000*HZ)/1000)

// Frequency list (map channels to frequencies)
const long frequency_list[] = { 2412, 2417, 2422, 2427, 2432, 2437, 2442,
				2447, 2452, 2457, 2462, 2467, 2472, 2484 };

typedef struct _net_local {
	  NET_DEVICE		*dev;		// backtrack device	
	  SEM_ADAPTER          adapter;   // main data structure
      spinlock_t		slock;
      struct net_device_stats	stats;	// device stats
//#ifdef WIRELESS_EXT
	  struct iw_statistics	wstats;	// wireless stats
//#endif /* WIRELESS_EXT */

} net_local;

#define PCI_VENDOR_ID_SEMCO    0x167D
#define PCI_DEVICE_ID_MINIPCI  0xA000

/********************************************************************
 * FUNCTION PROTOTYPES
 */
extern int init_module (void);
extern void cleanup_module (void);

static int sem_init_one(struct pci_dev *pdev,
		const struct pci_device_id *ent);
static void sem_remove_one (struct pci_dev *pdev);
//#ifdef CONFIG_SEM_PM
//static void sem_suspend (struct pci_dev *pdev);
//static void sem_resume (struct pci_dev *pdev);
//#endif
static int sem_found1(struct pci_dev *pdev,long ioaddr);
static struct net_device_stats *sem_get_stats (NET_DEVICE *dev);

//#ifdef WIRELESS_EXT
static int sem_ioctl (NET_DEVICE *dev, struct ifreq *rq, int cmd);
static struct iw_statistics *sem_get_wireless_stats (NET_DEVICE *dev);
//#endif /* WIRELESS_EXT */
static void sem_isr (int irq, void *dev_id, struct pt_regs *regs);
static int sem_tx (struct sk_buff *skb, NET_DEVICE *dev);
static WORD sem_rx (NET_DEVICE *dev);

static int sem_init (NET_DEVICE *dev);
static int sem_open (NET_DEVICE *dev);
static int sem_close (NET_DEVICE *dev);

int sem_hw_config (NET_DEVICE *dev);


static int sem_init_one (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{

	unsigned long ioaddr;
	int irq, i;
	int acpi_idle_state = 0, pm;
	static int cards_found ;

//	static int did_version;

	DEBUG(DEBUG_CALLTRACE, "-> sem_init_one()\n");

//	if (speedo_debug > 0  &&  did_version++ == 0)
//		printk(version);

	// save power state b4 pci_enable_device overwrites it
	pm = pci_find_capability(pdev, PCI_CAP_ID_PM);
	if (pm) {
		u16 pwr_command;
		pci_read_config_word(pdev, pm + PCI_PM_CTRL, &pwr_command);
		acpi_idle_state = pwr_command & PCI_PM_CTRL_STATE_MASK;
	}

	i = pci_enable_device(pdev);
	if (i){
     DEBUG(DEBUG_CALLTRACE, "<- sem_init_one()\n");
     return i;
    }

//	if (!request_region(pci_resource_start(pdev, 1),
//			pci_resource_len(pdev, 1), "swld12")) {
//		printk (KERN_ERR "swld12: cannot reserve I/O ports\n");
//		goto err_out_none;
//	}
	if (!request_mem_region(pci_resource_start(pdev, 0),
			pci_resource_len(pdev, 0), "swld12")) {
		printk (KERN_ERR "swld12: cannot reserve MMIO region\n");
		goto err_out_none;
	}

	irq = pdev->irq;

	ioaddr = (unsigned long)ioremap(pci_resource_start(pdev, 0),
									pci_resource_len(pdev, 0));
	if (!ioaddr) {
		printk (KERN_ERR "swld12: cannot remap MMIO region %lx @ %lx\n",
				pci_resource_len(pdev, 0), pci_resource_start(pdev, 0));
		goto err_out_free_mmio_region;
	}
//	if (speedo_debug > 2)
		printk("Found SAMSUNG MiniPCI/PCI, MMIO at %#lx(%lx),Virtual MMIO at %lx, IRQ %d\n",
			   pci_resource_start(pdev, 0),pci_resource_len(pdev, 0), ioaddr,irq);


	pci_set_master(pdev);

	if (sem_found1(pdev, ioaddr) == 0)
		cards_found++;
	else
		goto err_out_iounmap;

    DEBUG(DEBUG_CALLTRACE, "<- sem_init_one()\n");
	return 0;

err_out_iounmap:
	iounmap ((void *)ioaddr);
err_out_free_mmio_region:
	release_mem_region(pci_resource_start(pdev, 0), pci_resource_len(pdev, 0));
err_out_none:
    DEBUG(DEBUG_CALLTRACE, "<- sem_init_one()\n");
	return -ENODEV;

}

static int sem_found1(struct pci_dev *pdev,long ioaddr){
    NET_DEVICE *dev=0;    //struct net_device.
    net_local  *local;
//    BYTE i;

//    dev= init_etherdev(NULL,sizeof(net_local));
//    if(dev == NULL){
//      printk(KERN_ERR "sem12:Could not allocate ethernet device.\n");
//      return -1;
//     }
    //allocate space for net device
    dev = kmalloc(sizeof(NET_DEVICE),GFP_KERNEL);
    memset(dev,0,sizeof(NET_DEVICE));

    //allocate space for private area.
    local = kmalloc(sizeof(net_local),GFP_KERNEL);
    memset(local,0,sizeof(net_local));

    dev->priv = local;

   // local = dev->priv;
    dev->irq  = pdev->irq;
    dev->base_addr = ioaddr;

    //initialize specific data
    local->dev = dev;
    //standard setup for generic data
    ether_setup(dev);

    pdev->driver_data = dev;
    //kernel callback
    dev->init = &sem_init;
    dev->open = &sem_open;
    dev->stop = &sem_close;
    dev->hard_start_xmit = &sem_tx;
    dev->get_stats = &sem_get_stats;
    dev->do_ioctl = &sem_ioctl;
    dev->get_wireless_stats = &sem_get_wireless_stats;

    netif_stop_queue(dev);
    dev->mtu = mtu;
    strcpy(dev->name,"eth%d");
    sem_hw_config(dev);

    //register the net_device.
    register_netdev(dev);
    return 0;
}

int sem_hw_config (NET_DEVICE *dev)
//Description :
//      make the firmware into working condition.
//Return value:
//      0  : Success
	
{
	
	net_local *local = (net_local *) dev->priv;
	WORD  i,status = COMMAND_SUCCESS;
    SEM_ADAPTER *Adapter = &local->adapter;

	DEBUG(DEBUG_CALLTRACE, "-> sem_hw_config\n");
	
	Adapter->SystemStatus = NORMALMODE;
	//save the resource information into private data structure.
	DEBUG(DEBUG_CALLTRACE,"SWLD11 Driver Resource & Parameter Setting As to...\n");
    Adapter->ioaddr = dev->base_addr;
      DEBUG(DEBUG_CALLTRACE,"   IOBase = 0x%lx\n",Adapter->ioaddr);
    Adapter->InterruptNumber = dev->irq;
      DEBUG(DEBUG_CALLTRACE,"   IRQ = %d\n",Adapter->InterruptNumber);
    Adapter->NetworkMode  = networkmode;
      DEBUG(DEBUG_CALLTRACE,"   Network Mode = %d\n",Adapter->NetworkMode);
    //Adapter->CurrentRegDomain = Domain;
      //  DEBUG(DEBUG_CALLTRACE,"   Domain = %d\n",Adapter->CurrentRegDomain);
    Adapter->PSModeValue = psmode;
      DEBUG(DEBUG_CALLTRACE,"   PSMode = %d\n",Adapter->PSModeValue);
    Adapter->WepRequired = wepreq;
      DEBUG(DEBUG_CALLTRACE,"   WepReq = %d\n",Adapter->WepRequired);

    Adapter->WepDefaultKeyId = keyid;
      DEBUG(DEBUG_CALLTRACE,"   Key ID = %d\n",Adapter->WepDefaultKeyId);

    if(key0 != ""){
       semstring2hex(key0,Adapter->WepDefaultKey0);
       if(Adapter->WepRequired == 3)
            Adapter->WepDefaultKey0[5] = 0x00;
    }
    if(key1 != ""){
          semstring2hex(key1,Adapter->WepDefaultKey1);
		if(Adapter->WepRequired == 3)
            Adapter->WepDefaultKey1[5] = 0x00;
     }
    if(key2 != ""){
        semstring2hex(key2,Adapter->WepDefaultKey2);
		if(Adapter->WepRequired == 3)
            Adapter->WepDefaultKey2[5] = 0x00;
        }
    if(key3 != ""){
          semstring2hex(key3,Adapter->WepDefaultKey3);
		if(Adapter->WepRequired == 3)
            Adapter->WepDefaultKey3[5] = 0x00;
    }


        DEBUG(DEBUG_CALLTRACE,"   Input Key0 = %s\n",key0);
        DEBUG(DEBUG_CALLTRACE,"   Input Key1 = %s\n",key1);
        DEBUG(DEBUG_CALLTRACE,"   Input Key2 = %s\n",key2);
        DEBUG(DEBUG_CALLTRACE,"   Input Key3 = %s\n",key3);

    if(Adapter->WepRequired == WEP_128BIT){
         DEBUG(DEBUG_CALLTRACE,"   Key0= %x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x,%x\n",
							Adapter->WepDefaultKey0[0],Adapter->WepDefaultKey0[1],
                                   Adapter->WepDefaultKey0[2],Adapter->WepDefaultKey0[3],Adapter->WepDefaultKey0[4],
                                   Adapter->WepDefaultKey0[5],Adapter->WepDefaultKey0[6],Adapter->WepDefaultKey0[7],
							Adapter->WepDefaultKey0[8],Adapter->WepDefaultKey0[9],Adapter->WepDefaultKey0[10],
							Adapter->WepDefaultKey0[11],Adapter->WepDefaultKey0[12],Adapter->WepDefaultKey0[13]);
		}
    else{
          DEBUG(DEBUG_CALLTRACE,"   Key0= %x,%x,%x,%x,%x,%x\n",Adapter->WepDefaultKey0[0],Adapter->WepDefaultKey0[1],
                                   Adapter->WepDefaultKey0[2],Adapter->WepDefaultKey0[3],Adapter->WepDefaultKey0[4],
                                   Adapter->WepDefaultKey0[5]);
          DEBUG(DEBUG_CALLTRACE,"   Key1= %x,%x,%x,%x,%x,%x\n",Adapter->WepDefaultKey1[0],Adapter->WepDefaultKey1[1],

                                   Adapter->WepDefaultKey1[2],Adapter->WepDefaultKey1[3],Adapter->WepDefaultKey1[4],
                                   Adapter->WepDefaultKey1[5]);
          DEBUG(DEBUG_CALLTRACE,"   Key2= %x,%x,%x,%x,%x,%x\n",Adapter->WepDefaultKey2[0],Adapter->WepDefaultKey2[1],
                                   Adapter->WepDefaultKey2[2],Adapter->WepDefaultKey2[3],Adapter->WepDefaultKey2[4],
                                   Adapter->WepDefaultKey2[5]);
          DEBUG(DEBUG_CALLTRACE,"   Key3= %x,%x,%x,%x,%x,%x\n",Adapter->WepDefaultKey3[0],Adapter->WepDefaultKey3[1],
                                   Adapter->WepDefaultKey3[2],Adapter->WepDefaultKey3[3],Adapter->WepDefaultKey3[4],
                                   Adapter->WepDefaultKey3[5]);
       }
    Adapter->AuthenticationType = 1; //1: open system, 2: shared key

    Adapter->Channel      = channel;

    Adapter->APDensity    = apdensity;
         DEBUG(DEBUG_CALLTRACE,"   AP Density = %d\n",Adapter->APDensity);
    Adapter->Threshold    = threshold; //RTS/CTS Threshold
         DEBUG(DEBUG_CALLTRACE,"   RTS/CTS Threshold = %d\n",Adapter->Threshold);
    Adapter->FragThresh    = fragthresh;
         DEBUG(DEBUG_CALLTRACE,"   FragThreshold = %d\n",Adapter->FragThresh);
    Adapter->TransmitRate = txrate;
         DEBUG(DEBUG_CALLTRACE,"   Transmit Rate = %d\n",Adapter->TransmitRate);

    strcpy(Adapter->Desired_ESSID,essid);
    Adapter->Desired_ESSID_Length = strlen(essid);
         DEBUG(DEBUG_CALLTRACE,"   ESSID = %s\n",Adapter->Desired_ESSID);
    strcpy(Adapter->StationName,stationname);
    Adapter->StationNameLength = strlen(stationname);
         DEBUG(DEBUG_CALLTRACE,"   Sation Name = %s\n",Adapter->StationName);

	//commit the INITIALIZE command.

    Adapter->CardStat = CARD_STAT_PRESENT;
	status = Initialize(Adapter);
	if(status != COMMAND_SUCCESS){
	  DEBUG(DEBUG_CALLTRACE,"Initialize() failed...\n");
	  return status;
	}
     else
       DEBUG(DEBUG_CALLTRACE,"Initialize() Success...\n");

    spin_lock_init(&local->slock);
    //disable interrupt
    Adapter->IntEnState = 0;
    semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);

	//read the initial setting of Firmware ,just for debug purpose.
	//cardStatusRead(Adapter);

	//set initial value into firmware.
	cardInitialSet(Adapter);
	
	//save the mac address of the card into DEVICE structure.
       for(i=0;i<SEM_LENGTH_OF_ADDRESS;++i)
         dev->dev_addr[i] = Adapter->CurrentNetworkAddress[i];

     ////////////////////////////allocate the buffer for TX//////////////////////////////////
      // commit the ALLOCATE command.
      for(i=0;i<10;++i){
       status = issueCmd1(Adapter,ALLOCATE,MAX_DATALEN+60);

        if(status != COMMAND_SUCCESS){
         DEBUG(DEBUG_CALLTRACE,"--> Coomand ALLOCATE Error,status = %x\n",status);
         issueEventACK(Adapter,ALLOCACK);
         if(i==10)
           return status;
        }
        else break;
       } //end of for statement....

       //wait until the ALLOC event is happened.
       for(i=0;i<10;++i){
         status = waitevent(Adapter,ALLOC);
         if(status == COMMAND_SUCCESS) break;
       }
       if(status != COMMAND_SUCCESS){
         //DEBUG(DEBUG_CALLTRACE,"-->event ALLOC Error,status = %x\n",result);
         //local->stats.tx_dropped++;
         return status;
       }

	   semGetIOPortWord(Adapter,ALLOCFID,Adapter->FIDValue); // read AllocFID IO PORT(22h).
       DEBUG(DEBUG_CALLTRACE,"-->Alloc FID = 0x%x\n",Adapter->FIDValue);
       //issue ACK about ALLOC event.
       issueEventACK(Adapter,ALLOCACK);
       Adapter->reclaimStatus = TRUE;
     ////////////////////////////////////////////////////////////////////////////////////////////

	//commit the ENABLE command.
//    status = Enable(Adapter);
//	if(status != COMMAND_SUCCESS){
//	 DEBUG(DEBUG_CALLTRACE,"Enable() failed...\n");
//        return status;
//      }
	
	//read the setting value of firmware(for debugging..).
	//cardStatusRead(Adapter);
	
//#if (LINUX_VERSION_CODE >= VERSION(2,3,99))
//	netif_start_queue(dev);
///#endif
	//enable interrupt.
///////////	semPutIOPortWord(Adapter,INTEN,INFDROP | INFOEN | RXEN | ALLOCEN);


	DEBUG(DEBUG_CALLTRACE, "<- sem_hw_config()\n");
	
	return COMMAND_SUCCESS;
	
}

//======================================== INTERRUPT HANDLER ==========================================
static void sem_isr(int irq, void *dev_id, struct pt_regs *regs)
/**
 Routine Description:
   This Routine enables interrupts on the adapter.

 Arguments:
   Context - Really a  pointer to the adapter.

  Return Value:
    None

**/
{
// Holds the pointer the adapter.	
      NET_DEVICE *dev = (NET_DEVICE *) dev_id;
      net_local *local = (net_local *) dev->priv;
      int cnt;

      WORD Status,InfoFID;
      INTERRUPT_TYPE interruptStatus;
      WORD interruptState;
      SEM_ADAPTER *Adapter = &local->adapter;
//      DWORD flags;

//     DEBUG(DEBUG_INTERRUPT,"-->sem_isr\n");
     if(Adapter->CardStat){ // check the current Card status!.

      spin_lock(&local->slock);
      Adapter->IntEnState = 0;
      semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);

      semGetIOPortWord(Adapter,EVSTAT,interruptState);
      Adapter->InterruptStatus |= interruptState ;

      //semGetIOPortWord(Adapter,INTEN,IntEnState);
      interruptStatus = GetInterruptType(Adapter,Adapter->InterruptStatus);

      //check if it is ours or not...
      if(((interruptState & (INFDROP | INFOEN | RXEN | ALLOCEN)) == 0xFFFF) || ((interruptState & (INFDROP | INFOEN | RXEN | ALLOCEN)) == 0)){
            Adapter->IntEnState = INFDROP | INFOEN | RXEN | ALLOCEN;
		    semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);
            spin_unlock(&local->slock);
//			DEBUG(DEBUG_INTERRUPT,"-->It's NOT ours...\n");
            return;
      }

//      DEBUG(DEBUG_INTERRUPT,"sem_isr, Event  = %x\n",Adapter->InterruptStatus);
//#if (LINUX_VERSION_CODE < VERSION(2,3,99))
//      dev->interrupt = 1;
//#endif
      cnt = 20;
     while((interruptStatus  != UNKNOWN) && (Adapter->CardStat)&& (cnt > 0)){
      //DEBUG(DEBUG_INTERRUPT,"sem_isr, Event  = %x\n",Adapter->InterruptStatus);
      switch(interruptStatus){
     // ================================= RX interrupt Part ============================================
     case RX_INT:
       // RX interrupt.		
       //DEBUG_CALLTRACE(DEBUG_INTERRUPT,"Rx INT ...\n");
        //Adapter->InterruptStatus &= ~(RX);
	 //issueEventACK(Adapter,RXACK);
        Status = sem_rx(dev);
	    Adapter->InterruptStatus &= ~(RX);
	    issueEventACK(Adapter,RXACK);	
	    break;
      //================================= Tx Success Part ============================================
    case TX_INT:
	   Adapter->InterruptStatus &= ~(TX);
	   issueEventACK(Adapter,TXACK);	
	  break;
      //================================ TX Error Part ===============================================
   case TXEXC_INT:	
     //DEBUG(DEBUG_INTERRUPT,"Tx Error: ");
     Adapter->InterruptStatus &= ~(TXEXC);
     issueEventACK(Adapter,TXEXCACK);	
	   break;
   //-------------------- Info Part --------------------------------------
   case INFO_INT:
        //  DEBUG(DEBUG_INTERUPT,"  Info INT..\n");
     semGetIOPortWord(Adapter,INFOFID,InfoFID);
        // read the current statistic.
     Status = getInfoFrame(Adapter,InfoFID);
	if(Status != COMMAND_SUCCESS){
	  //DEBUG(DEBUG_CALLTRACE,"    Info Frame read error!\n");
        }
     Adapter->InterruptStatus &= ~(INFO);
	issueEventACK(Adapter,INFOACK);
    break;	
   case INFDROP_INT:	
       //DEBUG(DEBUG_INTERRUPT,"InfDrop event is happened\n");
	  Adapter->InterruptStatus &= ~(INFDROP);
	  issueEventACK(Adapter,INFDROPACK);
	  break;
   case ALLOC_INT:
        //DEBUG(DEBUG_INTERRUPT,"ALLOC event is happened\n");
//#if (LINUX_VERSION_CODE < VERSION(2,3,99))
//      dev->tbusy = 0;
//	 mark_bh(NET_BH);
//#else
	netif_wake_queue(dev);
//#endif
	 Adapter->reclaimStatus = TRUE;
	 Adapter->InterruptStatus &= ~(ALLOC);
     issueEventACK(Adapter,ALLOCACK);	
	  break;

   case TICK_INT:	
        //DEBUG(DEBUG_INTERRUPT,"-->SemHandleInterrupt,TickAck..\n");
	  Adapter->InterruptStatus &= ~(TICK);
	  issueEventACK(Adapter,TICKACK);
	  break;
   default:

	  break;
	   }// end of switch statement.

       //for improving the Rx Rate.
	 semGetIOPortWord(Adapter,EVSTAT,interruptState);
	 Adapter->InterruptStatus |= interruptState ;
		
       //Get the next interrupt to process.
    interruptStatus = GetInterruptType(Adapter,Adapter->InterruptStatus);

    --cnt;
     } //end of while.

     Adapter->IntEnState = INFDROP | INFOEN | RXEN | ALLOCEN;
     semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);
//#if (LINUX_VERSION_CODE < VERSION(2,3,99))
//     dev->interrupt = 0;
//#endif

     spin_unlock(&local->slock);

	 } // end of if(Adapter->CardStat....

//     DEBUG(DEBUG_INTERRUPT,"<-- sem_isr()\n\n");
}

static int sem_tx (struct sk_buff *skb, NET_DEVICE *dev)
// return value
//  0 - success, Nonzero = fail;
{
	
       net_local *local = (net_local *)dev->priv;
       //unsigned long flags;
//     int rc, len;
       BYTE *pBuffer;
       WORD result=0,Length,DataSended;
	   DWORD flags;

       SEM_ADAPTER *Adapter = &local->adapter;
      //DEBUG(DEBUG_TXRX, "-> sem_tx(%s)\n", dev->name);
      if(Adapter->CardStat){
       spin_lock_irqsave(&(local->slock),flags);

	    semPutIOPortWord(Adapter,INTEN,0);

	// Send packet
//#if (LINUX_VERSION_CODE >= VERSION(2,3,99))
	netif_stop_queue(dev);
//#else	
//	dev->tbusy = 1;
//#endif

	pBuffer = skb->data;
	Length = (ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN;
	 	
    if(Adapter->reclaimStatus == TRUE){
 	  result = sendTxFrame(Adapter,pBuffer,Length,&DataSended);
	  if(result == COMMAND_SUCCESS){
	
//#if (LINUX_VERSION_CODE >= VERSION(2,1,25))
	 local->stats.tx_bytes += DataSended;
//#endif
       //dev->trans_start = jiffies;	
       //if (!dev->tbusy)
	  // printk(KERN_WARNING "%s: Non-existent Transmission successful completed!?\n", dev_info);
	 //DEBUG(DEBUG_TXRX, "%s: Transmission successful completed\n", dev_info);
	 //dev->tbusy = 0;
	  // mark_bh(NET_BH);
	   local->stats.tx_packets++;
       }
       else{
        local->stats.tx_dropped++;
        }
       }
      else{
        DEBUG(DEBUG_TXRX, "reclaim is not happening...\n");
        local->stats.tx_dropped++;
 		result =1;
		}

 //      mark_bh(NET_BH);
 //      dev->tbusy = 0;
	  dev_kfree_skb(skb);

       // Re-enable interrupts
	
      semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);
      spin_unlock_irqrestore(&(local->slock),flags);
      } //end of if(Adapter->CardStat.....

	//DEBUG(DEBUG_TXRX, "<- sem_tx()\n\n");
	return result;

}

static WORD sem_rx (NET_DEVICE *dev)
{
	net_local *local = (net_local *)dev->priv;
	struct sk_buff *skb;
	BYTE *pBuffer,*bytepRxFrame;

     WORD DataLength=0,WordLength=0,skb_len=0;//,FrameControl=0,SeqControl=0;
	WORD result,FIDValue;

	WORD *pRxFrame;
//	rx_frame RxFrame;
	SEM_ADAPTER *Adapter = &local->adapter;
	
     //DEBUG(DEBUG_TXRX, "-> sem_rx(%s)\n", dev->name);

	pRxFrame = (WORD*)&(Adapter->RxFrame);
	bytepRxFrame = (BYTE*)&(Adapter->RxFrame);

    //get the received RX FID buffer.
	semGetIOPortWord(Adapter,RXFID,FIDValue);
    //read RX Frame MAC Header.
    result = getFrame0(Adapter,FIDValue,pRxFrame,0,30);
	
	if(result != COMMAND_SUCCESS){
         //DEBUG(DEBUG_CALLTRACE,"-->SemReceive(), Error Read RX header field field \n");
         return result;
	}

      //copy the RX Control data into adapter Structure for Statics.	
      Adapter->RXControl.Status  = (Adapter->RxFrame).r_rx_control.Status;
      //DEBUG(DEBUG_TXRX,"   Status0 = %x\n",Adapter->RXControl.Status);

      Adapter->RXControl.Signal  = (Adapter->RxFrame).r_rx_control.Signal;
      Adapter->RXControl.Silence = (Adapter->RxFrame).r_rx_control.Silence;
      Adapter->RXControl.RxFlow  = (Adapter->RxFrame).r_rx_control.RxFlow;
      Adapter->RXControl.Rate    = (Adapter->RxFrame).r_rx_control.Rate;

      DataLength = (Adapter->RxFrame).r_802_11_header.DataLen;
      //DEBUG(DEBUG_TXRX,"-->SemReceive: RX DataLen = %x\n",DataLength);

      if(DataLength >2304) DataLength = 0; // it means byte,,,,
      if(DataLength == 0){
         //DEBUG(DEBUG_CALLTRACE,"-->SemReceive,Data Length %x\n",DataLength);
	   return COMMAND_ERROR;
	  }
	
      // Create skb packet
     skb_len = DataLength + 14;
	skb = dev_alloc_skb(skb_len+2);  // it should accounts the room for 802.3 header(14byte).
	if (!skb)
	{
		printk(KERN_WARNING "%s Rx cannot allocate buffer for new packet\n",dev->name);
		local->stats.rx_dropped++;
		return COMMAND_ERROR;
	}

	// Align IP on 16b boundary
	skb_reserve(skb, 2);
     pBuffer = skb_put(skb, skb_len); // it should accounts the 802.3 header.
	dev->last_rx = jiffies;
	

/*
    Status = (Adapter->RxFrame).r_rx_control.Status;
     //DEBUG(DEBUG_TXRX,"   Status = %x\n",Status);
	
      //DEBUG(DEBUG_TXRX,"SEM--> Network DA addr %x %x %x %x %x %x\n",(Adapter->RxFrame).r_802_3_header.DestAddr[0],
			// (Adapter->RxFrame).r_802_3_header.DestAddr[1],(Adapter->RxFrame).r_802_3_header.DestAddr[2],
	   	     //  (Adapter->RxFrame).r_802_3_header.DestAddr[3],(Adapter->RxFrame).r_802_3_header.DestAddr[4],
			// (Adapter->RxFrame).r_802_3_header.DestAddr[5]);

      //DEBUG(DEBUG_TXRX,"SEM--> Network SA addr %x %x %x %x %x %x\n",(Adapter->RxFrame).r_802_3_header.SrcAddr[0],
			 //(Adapter->RxFrame).r_802_3_header.SrcAddr[1],(Adapter->RxFrame).r_802_3_header.SrcAddr[2],
		      // (Adapter->RxFrame).r_802_3_header.SrcAddr[3],(Adapter->RxFrame).r_802_3_header.SrcAddr[4],
			 //(Adapter->RxFrame).r_802_3_header.SrcAddr[5]);
*/
      //read rx frame the Data field.
      WordLength = (DataLength/2) + (DataLength%2);

      //read the all rx data.
      result = getFrame0(Adapter,FIDValue,pRxFrame,0,(WORD)(WordLength+30));
	if(result != COMMAND_SUCCESS){
       //DEBUG(DEBUG_CALLTRACE,"-->SemReceive(), Error Read Data Type \n");
       return result;
	 }
			
     result = semReceive(Adapter,pBuffer,&DataLength);

     if((result == COMMAND_SUCCESS) && (DataLength >0)){
	  skb->dev = dev;
	  skb->protocol = eth_type_trans(skb, dev);
	  skb->ip_summed = CHECKSUM_NONE;

	 // Hand the packet over to the kernel
	  netif_rx(skb);

	 local->stats.rx_packets++;
//#if (LINUX_VERSION_CODE >= VERSION(2,1,25))
	 local->stats.rx_bytes += DataLength;
//#endif

	// DEBUG(DEBUG_TXRX, "<- sem_rx()\n\n");

      return COMMAND_SUCCESS;
      }
    else return COMMAND_ERROR;
}


/********************************************************************
 * NET INIT / OPEN / CLOSE
*********************************************************************/
static void sem_remove_one (struct pci_dev *pdev)
{

	NET_DEVICE *dev = pdev->driver_data;
    net_local *local = (net_local *)dev->priv;
    DEBUG(DEBUG_CALLTRACE, "-> sem_remove_one()\n");
	
	unregister_netdev(dev);

	//release_region(pci_resource_start(pdev, 1), pci_resource_len(pdev, 1));
	release_mem_region(pci_resource_start(pdev, 0), pci_resource_len(pdev, 0));

	iounmap((char *)dev->base_addr);

//	pci_free_consistent(pdev, TX_RING_SIZE * sizeof(struct TxFD)
//								+ sizeof(struct speedo_stats),
//						sp->tx_ring, sp->tx_ring_dma);
    if(local) kfree(local);
    if(dev)   kfree(dev);
    printk(KERN_INFO "SAMSUNG 11Mbps WLAN MiniPCI/PCI card driver is removed\n\n");
}

static int sem_init (NET_DEVICE *dev)
{
	DEBUG(DEBUG_CALLTRACE, "<> sem_init()\n");
	return 0;
}

static int sem_open (NET_DEVICE *dev)
{
	net_local *local = (net_local*) dev->priv;
    SEM_ADAPTER *Adapter = &local->adapter;
    int retval;
    WORD status;

	DEBUG(DEBUG_CALLTRACE, "-> sem_open(%s)\n", dev->name);

	// TODO: Power up the card here and power down on close?
	// For now this is done on device init, not on open
	// Might be better placed here so that some settings can
	// be made by shutting down the device without removing
	// the driver (iwconfig).
	// But this is no real problem for now :-)

	// Start reception and declare the driver ready
	if (!Adapter->CardStat)
		return -ENODEV;
		
	MOD_INC_USE_COUNT;	
	
    //alloacate the ISR routine...
    retval = request_irq(dev->irq,&sem_isr,SA_SHIRQ | SA_INTERRUPT,dev->name,dev);
    if(retval){
      // Close the device
  	  MOD_DEC_USE_COUNT;
      return retval;
    }

	
///#if (LINUX_VERSION_CODE >= VERSION(2,3,99))
	netif_device_attach(dev);
	netif_start_queue(dev);
///#else
///	dev->interrupt = 0;
///	dev->start = 1;
///	dev->tbusy = 0;
///#endif

     //commit the ENABLE command.
    status = Enable(Adapter);
	if(status != COMMAND_SUCCESS){
	 DEBUG(DEBUG_CALLTRACE,"Enable() failed...\n");
        return status;
      }
	
	
	//enable interrupt.
	Adapter->IntEnState = INFDROP | INFOEN | RXEN | ALLOCEN;
    semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);

    printk(KERN_WARNING "SAMSUNG 11Mbps WLAN MiniPCI/PIC Card is working...\n");
	DEBUG(DEBUG_CALLTRACE, "<- sem_open()\n");
	return 0;
}

static int sem_close (NET_DEVICE *dev)
{
	net_local *local = (net_local *) dev->priv;
    SEM_ADAPTER *Adapter = &local->adapter;
    WORD status;

	DEBUG(DEBUG_CALLTRACE, "-> sem_close(%s)\n", dev->name);

	// Check if card is still present
///#if (LINUX_VERSION_CODE >= VERSION(2,3,99))
	if(netif_running(dev))
	{
		netif_stop_queue(dev);
		netif_device_detach(dev);
	}
///#else
///	if (dev->start)
///	{
///		dev->tbusy = 1;
///		dev->start = 0;
///		// TODO: Shutdown hardware
///	}
///#endif
    status = Disable(Adapter);
	if(status != COMMAND_SUCCESS){
	 DEBUG(DEBUG_CALLTRACE,"Disable() failed...\n");
      //  return status;
      }
    //disable interrupt.
    Adapter->IntEnState = 0;
    semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);

    free_irq(dev->irq,dev);

    // Close the device
	MOD_DEC_USE_COUNT;

	printk(KERN_WARNING "SAMSUNG 11Mbps WLAN MiniPCI/PIC Card is closed\n");
	DEBUG(DEBUG_CALLTRACE, "<- sem_close()\n");
	return -EINVAL;
}

/********************************************************************
 * NET STATS / IOCTL
 */

static struct net_device_stats *sem_get_stats (NET_DEVICE *dev)
{
    net_local *local = (net_local *) dev->priv;

	DEBUG(DEBUG_CALLTRACE, "<> sem_get_stats(%s)\n", dev->name);
	if(local)
 	 return(&local->stats);
 	else{
 	 DEBUG(DEBUG_CALLTRACE, "private area is not existed and so return null value.\n");
 	 return 0;
 	 }
}

//#ifdef WIRELESS_EXT
static int sem_ioctl (NET_DEVICE *dev, struct ifreq *rq, int cmd)
{
	net_local *local = (net_local *) dev->priv;
	struct iwreq *wrq = (struct iwreq *) rq;
    DWORD flags;
	int rc = 0,i;
    WORD value,Length,status,subcmd;
    SEM_ADAPTER *Adapter = &local->adapter;
    char buff[256];

	//DEBUG(DEBUG_CALLTRACE, "-> sem_ioctl(%s, cmd=0x%x)\n", dev->name, cmd);

	// Only super-user may write any setting
	if (IW_IS_SET(cmd) && cmd >= SIOCSIWNAME && !suser())
		return -EPERM;

	 //check the card status.
	 if(!Adapter->CardStat) return 0;
	 	
	// Disable interrupts
	// This prevents from being disturbed by an interrupt
	// during our fiddling with the NIC.
	
	spin_lock_irqsave(&(local->slock),flags);
    semPutIOPortWord(Adapter,INTEN,0);

	switch (cmd)
	{
		// Get name
		case SIOCGIWNAME:
			strcpy(wrq->u.name, "SAMSUNG WLAN");
			break;
			
		case SIOCDEVPRIVATE:
		     if (wrq->u.data.pointer)
			{
			    subcmd = wrq->u.data.flags;
			    switch(subcmd)
			    {
			     case SWL_UIL_NETWORKMODE_GET:
				// Get the network mode.
				  buff[0] = Adapter->NetworkMode;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				case SWL_UIL_NETWORKMODE_SET:
				//set the network mode.
				 if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,1);
				    Adapter->NetworkMode = buff[0];
                       }
				  break;
				case SWL_UIL_PSMODE_GET:
				 // Get the PS MODE.
				  buff[0] = Adapter->PSModeValue;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				case SWL_UIL_PSMODE_SET:
				//set the PS mode.
				 if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,1);
				    Adapter->PSModeValue = buff[0];
				    }

				  break;
				case SWL_UIL_VER:
				 //driver ver(major,minor),primary f/w ver(major,minor,variant),secondary f/w ver(").
				  buff[0] = SEM_DRIVER_MAJOR_VERSION;
				  buff[1] = SEM_DRIVER_MINOR_VERSION;
                      buff[2]  = SEM_DRIVER_VARIANT_VERSION;
				  buff[3] = Adapter->PRIMajorVersion;
				  buff[4] = Adapter->PRIMinorVersion;
				  buff[5] = Adapter->PRIVariant;
				  buff[6] = Adapter->STAMajorVersion;
				  buff[7] = Adapter->STAMinorVersion;
				  buff[8] = Adapter->STAVariant;
				  wrq->u.data.length = 9;
				  copy_to_user(wrq->u.data.pointer,buff,9);
				  break;
				case SWL_UIL_CURRENT_ADDRESS:
				 //get the Current station MAC Address.
				  wrq->u.data.length = SEM_LENGTH_OF_ADDRESS;
         			  copy_to_user(wrq->u.data.pointer,Adapter->CurrentNetworkAddress,SEM_LENGTH_OF_ADDRESS);
                      break;
				case SWL_UIL_LINKSTATUS:
				  // Get the Link Status.
				  buff[0] = Adapter->LinkStatus;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				
				case SWL_UIL_COMMSQUALITY:
				 //get the Communication Quality.
				  rc = getCommsQuality(Adapter, (WORD*)buff);
		      if(rc == COMMAND_ERROR) break;
		       	
		      wrq->u.data.length = 6;
				  copy_to_user(wrq->u.data.pointer,buff,6);
         // DEBUG(DEBUG_CALLTRACE, "CQ = %x, ASL = %x\n",(wrq->u.data.pointer[1]<<8)+ wrq->u.data.pointer[0],
				//					           (wrq->u.data.pointer[3]<<8)+ wrq->u.data.pointer[2]);
				  break;
				case SWL_UIL_TXRATE_GET:
				  //get the TX Rate.
				  buff[0] = (char)Adapter->TransmitRate;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				case SWL_UIL_TXRATE_SET:
				   //set the TX Rate.
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,1);
				    Adapter->TransmitRate = buff[0];
				    }
				
				  break;
				case SWL_UIL_WEPFLAGS_GET:
				   //get the WEP Status
				  buff[0] = (char)Adapter->WepRequired;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				case SWL_UIL_WEPFLAGS_SET:
				   //set the WEP Status.
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,1);
				    Adapter->WepRequired = buff[0];
				    }
				
				  break;
				case SWL_UIL_DOMAIN:
				   //get the domain
				  buff[0] = (char)Adapter->CurrentRegDomain;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				
				case SWL_UIL_WEPKEYID_GET:
				   //get the WEP Default KEy ID
				  buff[0] = (char)Adapter->WepDefaultKeyId;
				  wrq->u.data.length = 1;
				  copy_to_user(wrq->u.data.pointer,buff,1);
				  break;
				case SWL_UIL_WEPKEYID_SET:
				   //set the WEP Default KEy ID
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,1);
				    Adapter->WepDefaultKeyId = buff[0];
			        }
				  break;
                    case SWL_UIL_WEPKEY1_GET:
 				    if(Adapter->WepRequired == 7){	
					 wrq->u.data.length = 13;
				      copy_to_user(wrq->u.data.pointer,Adapter->WepDefaultKey0,13);
					}
                        else{
 					 wrq->u.data.length = 5;
                          copy_to_user(wrq->u.data.pointer,Adapter->WepDefaultKey0,5);
					}
				   break;
				case SWL_UIL_WEPKEY1_SET:
				  if(wrq->u.data.pointer){
                        if(Adapter->WepRequired == 7)
				      copy_from_user(buff,wrq->u.data.pointer,26);
                        else
				      copy_from_user(buff,wrq->u.data.pointer,10);

				    semstring2hex(buff,Adapter->WepDefaultKey0);

				    }
				  break;
			     case SWL_UIL_WEPKEY2_GET:
			        wrq->u.data.length = 5;
                       copy_to_user(wrq->u.data.pointer,Adapter->WepDefaultKey1,5);
				   break;
				case SWL_UIL_WEPKEY2_SET:
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,10);
				    semstring2hex(buff,Adapter->WepDefaultKey1);
				    }
				  break;
			     case SWL_UIL_WEPKEY3_GET:
				   wrq->u.data.length = 5;
                       copy_to_user(wrq->u.data.pointer,Adapter->WepDefaultKey2,5);
				   break;
				case SWL_UIL_WEPKEY3_SET:
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,10);
				    semstring2hex(buff,Adapter->WepDefaultKey2);
				    }
				  break;
  				case SWL_UIL_WEPKEY4_GET:
				   wrq->u.data.length = 5;
                       copy_to_user(wrq->u.data.pointer,Adapter->WepDefaultKey3,5);
				   break;
				case SWL_UIL_WEPKEY4_SET:
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,10);
				    semstring2hex(buff,Adapter->WepDefaultKey3);
				    }
				  break;
				case SWL_UIL_RTSCTS_GET:
				   //get the RTS/CTS Status
				  buff[0] = Adapter->Threshold & 0xFF;
				  buff[1] = (Adapter->Threshold>>8) & 0xFF;
				  wrq->u.data.length = 2;
				  copy_to_user(wrq->u.data.pointer,buff,2);
				  break;
				case SWL_UIL_RTSCTS_SET:
				   //set the RTS/CTS Status.
				  if(wrq->u.data.pointer){
				    copy_from_user(buff,wrq->u.data.pointer,2);
				    Adapter->Threshold = ((buff[1]&0xFF)<<8) + (buff[0]&0xFF);
				    }
				
				  break;
        		     case SWL_UIL_SITESURVEY_START:
					DEBUG(DEBUG_INFO, "start the SITESURVEY\n");
					siteSurvey_start(Adapter);	
				     break;
                    case SWL_UIL_SITESURVEY_END:
					DEBUG(DEBUG_INFO, "end the SITESURVEY\n");
					siteSurvey_end(Adapter);						
				     break;
                    case SWL_UIL_SITESURVEY_CONT:
					//DEBUG(DEBUG_INFO, "continue the SITESURVEY\n");
					//siteSurvey_cont(Adapter);					
                         break;
       	         case SWL_UIL_SCANRESULT:				
                         status = issueCmd1(Adapter,INQUIRE,SCANRESULTS);//cardUpdateScanResult(Adapter);
		               if(status != COMMAND_SUCCESS){
                          DEBUG(DEBUG_INFO, "query SCAN Result error!..\n");
                          break;
                         }
 	                    else{
                           //DEBUG(DEBUG_INFO, "length = %d\n",Length);
				       copy_to_user(wrq->u.data.pointer,Adapter->scan_buffer,
						         1000);//Lenght);
  					 }
                         break;

	              case SWL_UIL_SOFTRESET:
		             // software reset.
		              rc  = cardReset(Adapter);
		              break;
		       	
                   case SWL_UIL_ESSID:
                        status = getESSID(Adapter,(BYTE*)&buff,&Length);
             			if(status == COMMAND_ERROR) break;
                         wrq->u.data.length = Length;
                         buff[Length] = '\0';
				     copy_to_user(wrq->u.data.pointer,buff,Length+1);
                        break;
		         default:
			         rc = -EOPNOTSUPP;
			         break;
			      }
                       }//end of switch statement....
			break;
				
		// Set frequency/channel
		case SIOCSIWFREQ:
			// If setting by frequency, convert to a channel
			if((wrq->u.freq.e == 1) &&
			   (wrq->u.freq.m >= (int) 2.412e8) &&
			   (wrq->u.freq.m <= (int) 2.487e8))
			{
				int f = wrq->u.freq.m / 100000;
				int c = 0;
				while((c < 14) && (f != frequency_list[c]))
					c++;
				// Hack to fall through...
				wrq->u.freq.e = 0;
				wrq->u.freq.m = c + 1;
			}
			
			// Setting by channel number
			//if ((NetworkMode != 3) || (wrq->u.freq.m > 1000) || (wrq->u.freq.e > 0))
			//	rc = -EOPNOTSUPP;
			//else
			//{
               if(Adapter->NetworkMode !=1){
				value = wrq->u.freq.m;
				if (value<1 || value>15)
				{
					DEBUG(DEBUG_INFO, "New channel value of %d for %s is invalid!\n", wrq->u.freq.m, dev->name);
					rc = -EINVAL;

				}
				else
				{
					Adapter->Channel = value;					
					channel = value;
				}
               cardReset(Adapter);
 			  }
			//}

			break;

		// Get frequency/channel
		case SIOCGIWFREQ:
             status = getCurrentChannel(Adapter,&value);
             if(status == COMMAND_ERROR) break;
			 wrq->u.freq.m = value;
			 wrq->u.freq.e = 0;
			 break;

//#if WIRELESS_EXT > 5
		// Set desired network name (ESSID)
		case SIOCSIWESSID:
			if (wrq->u.data.pointer)
			{
				// Check if we asked for `any'
				if(wrq->u.data.flags == 0)
				{
					Adapter->Desired_ESSID[0]= '\0';
					Adapter->Desired_ESSID_Length = 0;
				}
				else
				{
					// Check the size of the string
					if(wrq->u.data.length > ESSID_MAX_SIZE + 1)
					{
						rc = -E2BIG;
						break;
					}
					copy_from_user(Adapter->Desired_ESSID,
						       wrq->u.data.pointer,
						       wrq->u.data.length);
				     Adapter->Desired_ESSID_Length = wrq->u.data.length-1;		
					Adapter->Desired_ESSID[ESSID_MAX_SIZE] = '\0';	
				}
				
			strncpy(essid, Adapter->Desired_ESSID, sizeof(essid)-1);
			DEBUG(DEBUG_CALLTRACE, "-> SET Essid =%s,Length=%d\n", Adapter->Desired_ESSID,Adapter->Desired_ESSID_Length);
			cardReset(Adapter);
			}
			break;

		// Get current network name (ESSID)
		case SIOCGIWESSID:
			if (wrq->u.data.pointer)
			{
				char essid1[ESSID_MAX_SIZE + 1];
				// Get the essid that was set
				status = getCurrentESSID(Adapter, essid1,&Length);
			       if(status == COMMAND_ERROR) break;
				// Push it out !
				wrq->u.data.length = Length;//strlen(essid) + 1;
				wrq->u.data.flags = 1; //active
				copy_to_user(wrq->u.data.pointer, essid1, sizeof(essid1));
			}
			break;

		// Get current Access Point (BSSID)
		case SIOCGIWAP:
		       if ((Adapter->NetworkMode != 3) &&
	              ((Adapter->LinkStatus  == CONNECTED)||
                  (Adapter->LinkStatus  == AP_IN_RANGE) ||
	              (Adapter->LinkStatus  == AP_CHANGE)))
	              {
      		       status = getCurrentBSSID(Adapter, wrq->u.ap_addr.sa_data);
		          if(status == COMMAND_ERROR) break;
		          }
		         else{
		           for(i=0;i<6;++i)
		             wrq->u.ap_addr.sa_data[i] = 0xFF;
		         }
			   wrq->u.ap_addr.sa_family = ARPHRD_ETHER;
			  break;

//#endif	// WIRELESS_EXT > 5

//#if WIRELESS_EXT > 7
		// Set desired station name
		case SIOCSIWNICKN:
			if (wrq->u.data.pointer)
			{
				// Check the size of the string
				if(wrq->u.data.length > ESSID_MAX_SIZE + 1)
				{
					rc = -E2BIG;
					break;
				}
				
				copy_from_user(Adapter->StationName, wrq->u.data.pointer, wrq->u.data.length);
				Adapter->StationName[ESSID_MAX_SIZE] = '\0';
				Adapter->StationNameLength = wrq->u.data.length;
				//cardReset(Adapter);
				strncpy(stationname, Adapter->StationName, sizeof(stationname)-1);
			}
			break;

		// Get current station name
		case SIOCGIWNICKN:
			if (wrq->u.data.pointer)
			{
				char name[ESSID_MAX_SIZE + 1];
				
				status = getStationName(Adapter, name,&Length);
			       if(status == COMMAND_ERROR) break;
				// Push it out !
				wrq->u.data.length = Length;
				wrq->u.data.flags = 1; //active				
				copy_to_user(wrq->u.data.pointer, name, sizeof(name));
			}
			break;
		// Set the desired bit-rate
		case SIOCSIWRATE:
			{
                /// int brate = wrq->u.bitrate.value/1000000;
                 int brate = wrq->u.bitrate.value;

				// Auto
				if(wrq->u.bitrate.fixed == 0)
				  Adapter->TransmitRate= 0x0F;	
				else{
					if(brate == 1) Adapter->TransmitRate = 1;
                    else if(brate == 2) Adapter->TransmitRate = 2;
					else if(brate == 5) Adapter->TransmitRate = 4;
				    else if(brate == 11) Adapter->TransmitRate = 8;
					else  Adapter->TransmitRate = 0x0F;
				}
				cardReset(Adapter);	
			}

			break;
		// Get the current bit-rate
		case SIOCGIWRATE:
            //if(Adapter->NetworkMode == 1){
			//  status = getCurrentTxRate(Adapter,&value);
        	//  if(status == COMMAND_ERROR) break;
            // }
            //
            // else{
              status = getTxRate(Adapter,&value);
        	  if(status == COMMAND_ERROR) break;
			//}
			
			if(value == 1) value =1; //1Mbps
			else if(value == 2) value =2; //2Mbps
            else if(value == 4) value =5; //5.5Mbps
            else if(value == 8) value = 11;//11Mbps
            else  value = 11;//11Mbps

			wrq->u.bitrate.value = value * 1000000;
		
			break;

		// Set the desired RTS threshold
		case SIOCSIWRTS:
		       {
				int rthr = wrq->u.rts.value;
				if(rthr == -1)
					rthr = 2347;
				if((rthr < 0) || (rthr > 2347))
				{
					rc = -EINVAL;
					break;
				}
				Adapter->Threshold = rthr;
				//cardReset(Adapter);
				threshold = rthr;
		       }
                cardReset(Adapter);	
			break;

		// Get the current RTS threshold
		case SIOCGIWRTS:
			wrq->u.rts.value = Adapter->Threshold;
			wrq->u.rts.fixed = 1;
			break;


		// Set the desired fragmentation threshold
		case SIOCSIWFRAG:
			{
				int fthr = wrq->u.frag.value;
				if(fthr == -1)
					fthr = 2346;
				if((fthr < 256) || (fthr > 2346) ||((fthr % 2) == 1))
				{
					rc = -EINVAL;
					break;
				}
				Adapter->FragThresh = fthr;
				fragthresh = fthr;
			}
               cardReset(Adapter);
			break;

		// Get the current fragmentation threshold
		case SIOCGIWFRAG:
			wrq->u.frag.value = Adapter->FragThresh;
			wrq->u.frag.fixed = 1;
			break;

		// Set the desired AP density
		case SIOCSIWSENS:
			{
				int dens = wrq->u.sens.value;
				if((dens < 1) || (dens > 3))
 				{
					rc = -EINVAL;
					break;
				}
				Adapter->APDensity = dens;
				//cardReset(Adapter);
				apdensity = dens;
			}
			break;

		// Get the current AP density
		case SIOCGIWSENS:
			status = getAPDensity(Adapter,&value);
      			if(status == COMMAND_ERROR) break;
			wrq->u.sens.value = value;
			wrq->u.sens.fixed = 0;	// auto
			break;
//#endif	// WIRELESS_EXT > 7

//#if WIRELESS_EXT > 8
		// Set port type
		case SIOCSIWMODE:
				switch (wrq->u.mode)
				{
					case IW_MODE_ADHOC:
						//Adapter->NetworkMode = 3;
						Adapter->NetworkMode = 0; // 802.11 ad_hoc
						cardReset(Adapter);
						break;
						// Fall through
					case IW_MODE_INFRA:
						Adapter->NetworkMode = 1;
						cardReset(Adapter);
                        break;
					default:
						rc = -EINVAL;
				}

			break;

		// Get port type
		case SIOCGIWMODE:
			 if(Adapter->NetworkMode == 1) wrq->u.mode = IW_MODE_INFRA;
			 else wrq->u.mode = IW_MODE_ADHOC;

			break;

		// Set the desired Power Management mode
		case SIOCSIWPOWER:
			// Disable it ?
			if(wrq->u.power.disabled)
				Adapter->PSModeValue = 0;
			else
			{
			 Adapter->PSModeValue = 1;
            }
			cardReset(Adapter);	
			break;

		// Get the power management settings
		case SIOCGIWPOWER:
			if(Adapter->PSModeValue == 0)
  			  wrq->u.power.disabled = 1;
 			else{	
			 wrq->u.power.disabled = 0;
			 wrq->u.power.flags = 0;
			 wrq->u.power.value = 0;
			 wrq->u.power.flags |= IW_POWER_ALL_R;
			 }
			break;

		// Set WEP keys and mode
		case SIOCSIWENCODE:
			// Basic checking: do we have a key to set?
			
			 // Set WepRequired according to Key Length: Michael Leun <ml@lara.leun.net> 12082001
		  if (wrq->u.encoding.length != 0)
		  {
		    if (wrq->u.encoding.length>5)
		      Adapter->WepRequired = 7;
		    else
		      Adapter->WepRequired = 3;
		  }
		
           // DEBUG(DEBUG_CALLTRACE, "Wep Required:%d\n", Adapter->WepRequired);
			if (wrq->u.encoding.pointer != (caddr_t) 0)
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				// Check the size of the key
				//if (wrq->u.encoding.length > MAX_KEY_SIZE)
				//{
				//	rc = -EINVAL;
				//	break;
				//}
				// Check the index
				if ((index < 0) || (index >= 4))
					index = Adapter->WepDefaultKeyId;
                    if(Adapter->WepRequired == 7) index =0;

				Adapter->WepDefaultKeyId = index;
				// Cleanup
				//memset(local->key[index].key, 0, MAX_KEY_SIZE);

				// Copy the key in the driver
			     if(Adapter->WepRequired == 7)
				      copy_from_user(Adapter->WepDefaultKey0,wrq->u.data.pointer,13);
                    else{
				   if(Adapter->WepDefaultKeyId == 0)
				      copy_from_user(Adapter->WepDefaultKey0,wrq->u.data.pointer,5);
				   else if(Adapter->WepDefaultKeyId == 1)
				      copy_from_user(Adapter->WepDefaultKey1,wrq->u.data.pointer,5);
				   else if(Adapter->WepDefaultKeyId == 2)
				      copy_from_user(Adapter->WepDefaultKey2,wrq->u.data.pointer,5);
				   else if(Adapter->WepDefaultKeyId == 3)
				      copy_from_user(Adapter->WepDefaultKey3,wrq->u.data.pointer,5);
			        else
				      copy_from_user(Adapter->WepDefaultKey0,wrq->u.data.pointer,5);
				}

			}
			else
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				// Do we want to just set the current transmit key?
				if ((index >= 0) && (index < 4))
				{
//					if (local->key[index].len > 0)
//					{
						Adapter->WepDefaultKeyId = index;
						Adapter->WepRequired |= 0x03;
//					}
//					else
//						rc = -EINVAL;
				}
			}
			// Read the flags
			if (wrq->u.encoding.flags & IW_ENCODE_DISABLED)
				Adapter->WepRequired = 0x00;	// disable encryption
               else
                    Adapter->WepRequired |= 0x03;	// enable encryption
              // DEBUG(DEBUG_CALLTRACE, "Wep Required:%d\n", Adapter->WepRequired);
			// Commit the changes
			if (rc == 0) {
				cardReset(Adapter);	
			}
			break;

		// Get the WEP keys and mode
		case SIOCGIWENCODE:
          DEBUG(DEBUG_CALLTRACE, "Wep Required:%d\n", Adapter->WepRequired);
			// Only super-user can see WEP key
			if (!capable(CAP_NET_ADMIN))
			{
				rc = -EPERM;
				break;
			}
			// Basic checking...
			if (wrq->u.encoding.pointer != (caddr_t) 0)
			{
				int index = (wrq->u.encoding.flags & IW_ENCODE_INDEX) - 1;
				// Note: should read from adapter(?), and check if WEP capable
				// Set the flags
				wrq->u.encoding.flags = 0;
				if (Adapter->WepRequired == 0)
					wrq->u.encoding.flags |= IW_ENCODE_DISABLED;
				// Which key do we want
				if ((index < 0) || (index >= 4))
					index = Adapter->WepDefaultKeyId;
				wrq->u.encoding.flags |= index ;
    				wrq->u.encoding.flags |= IW_ENCODE_RESTRICTED ;
				// Copy the key to the user buffer
                    if(Adapter->WepRequired == 7)
				 wrq->u.encoding.length = 13;//local->key[index].len;
				else
				 wrq->u.encoding.length = 5;
                    if(index == 0)
  				  copy_to_user(wrq->u.encoding.pointer, Adapter->WepDefaultKey0, wrq->u.encoding.length);
                    else if(index == 1)
  				  copy_to_user(wrq->u.encoding.pointer, Adapter->WepDefaultKey1, 5);
				else if(index == 2)
  				  copy_to_user(wrq->u.encoding.pointer, Adapter->WepDefaultKey2, 5);
				else if(index == 3)
  				  copy_to_user(wrq->u.encoding.pointer, Adapter->WepDefaultKey3, 5);
				else
  				  copy_to_user(wrq->u.encoding.pointer, Adapter->WepDefaultKey0, 5);

			}
			break;

//#endif // WIRELESS_EXT > 8

		// Get range of parameters
		case SIOCGIWRANGE:
			if (wrq->u.data.pointer)
			{
				struct iw_range range;
				rc = verify_area(VERIFY_WRITE, wrq->u.data.pointer, sizeof(struct iw_range));
				if (rc)
					break;
				wrq->u.data.length = sizeof(range);
				range.throughput = 11.0 * 1024 * 1024;	// 11 Mbps
				range.min_nwid = 0x0000;
				range.max_nwid = 0x0000;
				range.num_channels = 14;
				range.num_frequency = 14;//wvlan_hw_getfrequencylist(&local->ifb,
						     // range.freq,
						     // IW_MAX_FREQUENCIES);
				range.sensitivity = 3;
				if (Adapter->NetworkMode == 3)//Ad-hoc ?  //&& local->spy_number == 0)
				{
					range.max_qual.qual = 0;
					range.max_qual.level = 0;
					range.max_qual.noise = 0;
				}
				else
				{
					range.max_qual.qual = 0x8b - 0x2f;
					range.max_qual.level = 0x2f - 0x95 - 1;
					range.max_qual.noise = 0x2f - 0x95 - 1;
				}
//#if WIRELESS_EXT > 7
				range.num_bitrates = 4;//wvlan_getbitratelist(&local->ifb,
						//	range.bitrate,
						//	IW_MAX_BITRATES);
				range.min_rts = 0;
				range.max_rts = 2347;
				range.min_frag = 256;
				range.max_frag = 2346;
//#endif	// WIRELESS_EXT > 7
				copy_to_user(wrq->u.data.pointer, &range, sizeof(struct iw_range));
			}
			break;


		// All other calls are currently unsupported
		default:
			rc = -EOPNOTSUPP;
	}

    //restore the previous interrupt status.
    semPutIOPortWord(Adapter,INTEN,Adapter->IntEnState);
    spin_unlock_irqrestore(&(local->slock),flags);

	//DEBUG(DEBUG_CALLTRACE, "<- sem_ioctl()\n");
	return rc;
}


static struct iw_statistics *sem_get_wireless_stats (NET_DEVICE *dev)
{
	net_local *local = (net_local *) dev->priv;
	WORD status,buf[4];
    DWORD flags;
    SEM_ADAPTER *Adapter = &local->adapter;

	DEBUG(DEBUG_CALLTRACE, "-> sem_get_wireless_stats(%s)\n", dev->name);
	
	local->wstats.status = 0;
	if ((Adapter->NetworkMode != 3) &&
	    ((Adapter->LinkStatus  == CONNECTED)||
         (Adapter->LinkStatus  == AP_IN_RANGE) ||
	     (Adapter->LinkStatus  == AP_CHANGE)))
	{
	    spin_lock_irqsave(&(local->slock),flags);
		status= getCommsQuality(Adapter,buf);
		spin_unlock_irqrestore(&(local->slock),flags);
      if(status != COMMAND_ERROR){
		  local->wstats.qual.qual = buf[0];//max(min(ltv.coms_qual, 0x8b-0x2f), 0);
		  if(buf[1]>100) buf[1] -= 10;
		  local->wstats.qual.level = buf[1];//max(min(ltv.signal_lvl, 0x8a), 0x2f) - 0x95;
		  local->wstats.qual.noise = buf[2];//max(min(ltv.noise_lvl, 0x8a), 0x2f) - 0x95;
		  local->wstats.qual.updated = 7;
     }
      else{
                local->wstats.qual.qual = 0;
		  local->wstats.qual.level = 0;
		  local->wstats.qual.noise = 0;
		  local->wstats.qual.updated = 0;
      }
	}
	else
	{
		// Quality levels cannot be determined in ad-hoc mode,
		// because we can 'hear' more that one remote station.
		// If a spy address is defined, we report stats of the
		// first spy address
		local->wstats.qual.qual = 0;
		local->wstats.qual.level = 0;
		local->wstats.qual.noise = 0;
		local->wstats.qual.updated = 0;
	}

	// Packets discarded in the wireless adapter due to wireless specific problems
	local->wstats.discard.nwid = 0;
	local->wstats.discard.code = 0;//local->ifb.IFB_NIC_Tallies.RxWEPUndecryptable;
	local->wstats.discard.misc = 0;//local->ifb.IFB_NIC_Tallies.RxFCSErrors +
					//local->ifb.IFB_NIC_Tallies.RxDiscards_NoBuffer +
					//local->ifb.IFB_NIC_Tallies.TxDiscardsWrongSA;

	DEBUG(DEBUG_CALLTRACE, "<- sem_get_wireless_stats(%s)\n",dev->name);
	return (&local->wstats);
}


//#endif // WIRELESS_EXT
//********************************************************************

//********************************************************************
// * MODULE INSERTION / REMOVAL
//*********************************************************************/
static struct pci_device_id sem_pci_tbl[] = {
	{ PCI_VENDOR_ID_SEMCO, PCI_DEVICE_ID_MINIPCI,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,}
};
MODULE_DEVICE_TABLE(pci, sem_pci_tbl);
	
static struct pci_driver swld12_driver = {
	name:		"swld12",
	id_table:	sem_pci_tbl,
	probe:		sem_init_one,
	remove:		sem_remove_one,
//#ifdef CONFIG_SEM_PM
//	suspend:	sem_suspend,
//	resume:		sem_resume,
//#endif
};
extern int init_module (void)
{
//	servinfo_t serv;

	DEBUG(DEBUG_CALLTRACE, "-> init_module()\n");

	printk(KERN_INFO "SAMSUNG 11Mbps WLAN MiniPCI/PCI driver v%s\n",version);
	printk(KERN_INFO "Copyright(c) SAMSUNG Electro-Mechanics Co.,Ltd.\n");

	DEBUG(DEBUG_CALLTRACE, "<- init_module()\n");
	return pci_module_init(&swld12_driver);
}

extern void cleanup_module (void)
{
	DEBUG(DEBUG_CALLTRACE, "-> cleanup_module()\n");

	printk(KERN_INFO "SAMSUNG MiniPCI/PCI Driver unloaded\n");
	DEBUG(DEBUG_CALLTRACE, "<- cleanup_module()\n");

    pci_unregister_driver(&swld12_driver);
    DEBUG(DEBUG_CALLTRACE, "<- ####################cleanup_module()####################################\n");

}

